from pathlib import Path
from datetime import datetime, timezone
import pytz
import inspect
from slugify import slugify
import sys
import json
import socket
from urllib.parse import urlparse
import git

class ClearHelpers:
	@staticmethod
	#pattern:str = "FOXCAR.ARP.POSPAY1.{DT:EST:%Y%m%d_%H%M%S}.DATAFILE"
	def ReplaceTimestamp(pattern:str) -> str:
		returnValue:str = ""
		if ("{DT" in pattern):
			beginIndex:int = pattern.find("{DT")
			endIndex:int = pattern.find("}", beginIndex)
			if (beginIndex != (-1) and endIndex != (-1)):
				prefix:str = pattern[:beginIndex]
				suffix:str = pattern[endIndex+1:]
				dtStatement:str = pattern[beginIndex+1:endIndex]
				dtElements:list[str] = dtStatement.split(":")
				timeFormat:str = "%Y%m%d%H%M%S"
				useTimezone:timezone = datetime.now().astimezone().tzinfo
				if (len(dtElements) == 2):
					timeFormat = dtElements[1]
				elif (len(dtElements) == 3):
					match (dtElements[1]):
						case "L" | "l":
							useTimezone:timezone = datetime.now().astimezone().tzinfo
						case "Z" | "z" | "UTC" | "utc":
							useTimezone:timezone = timezone.utc
						case _:
							useTimezone:timezone = pytz.timezone(dtElements[1])
					timeFormat = dtElements[2]
				timestamp:str  = datetime.now(useTimezone).strftime(timeFormat)
				returnValue = f"{prefix}{timestamp}{suffix}"
		return returnValue

	@staticmethod
	#pattern:str = "FOXCAR.ARP.POSPAY1.{DT:%Y%m%d_%H%M%S}.DATAFILE"
	def ReplaceTimestampWithDateTime(pattern:str, dt:datetime) -> str:
		returnValue:str = ""
		if ("{DT" in pattern):
			beginIndex:int = pattern.find("{DT")
			endIndex:int = pattern.find("}", beginIndex)
			if (beginIndex != (-1) and endIndex != (-1)):
				prefix:str = pattern[:beginIndex]
				suffix:str = pattern[endIndex+1:]
				dtStatement:str = pattern[beginIndex+1:endIndex]
				dtElements:list[str] = dtStatement.split(":")
				timeFormat:str = "%Y%m%d%H%M%S"
				useTimezone:timezone = dt.astimezone().tzinfo
				if (len(dtElements) == 2):
					timeFormat = dtElements[1]
				elif (len(dtElements) == 3):
					match (dtElements[1]):
						case "L" | "l":
							useTimezone:timezone = datetime.now().astimezone().tzinfo
						case "Z" | "z" | "UTC" | "utc":
							useTimezone:timezone = timezone.utc
						case _:
							useTimezone:timezone = pytz.timezone(dtElements[1])
					timeFormat = dtElements[2]
				if (useTimezone != dt.astimezone().tzinfo):
					timestamp:str  = dt.astimezone(useTimezone).strftime(timeFormat)
				else:
					timestamp:str  = dt.strftime(timeFormat)
				returnValue = f"{prefix}{timestamp}{suffix}"
		return returnValue

	@staticmethod
	def ConvertToUTCEvenIfNaive(localized:datetime) -> datetime:
		returnValue:datetime = None
		if (localized.tzinfo == timezone.utc):
			returnValue = localized
		else:
			aware:datetime = None
			if (localized.tzinfo is not None):
				aware = localized
			else:
				aware = datetime(
					year=localized.year,
					month=localized.month,
					day=localized.day,
					hour=localized.hour,
					minute=localized.minute,
					second=localized.second,
					microsecond=localized.microsecond,
					tzinfo=datetime.now().astimezone().tzinfo
				)
			returnValue = aware.astimezone(timezone.utc)
		return returnValue

	@staticmethod
	def GetPathPartsFromRight(path:Path|str, numberOfElements:int, separator:str = "/", removeFileExtension:bool = False) -> str:
		returnValue:str = ""
		elementIndex:int = 0
		if (not isinstance(path, Path)):
			path = Path(path)
		if (removeFileExtension):
			path = path.with_suffix("")
		for element in Path(path).parts[::-1]:
			elementIndex += 1
			if (elementIndex == 1 and elementIndex <= numberOfElements):
				returnValue = element
			else:
				returnValue = f"{element}{separator}{returnValue}"
			if (elementIndex >= numberOfElements):
				break
		return returnValue

	@staticmethod
	def GetPathPartsFromLeft(path:str, numberOfElements:int, separator:str = "/") -> str:
		returnValue:str = ""
		elementIndex:int = 0
		for element in Path(path).parts:
			elementIndex += 1
			if (elementIndex == 1 and elementIndex <= numberOfElements):
				returnValue = element
			else:
				returnValue = f"{returnValue}{separator}{element}"
			if (elementIndex >= numberOfElements):
				break
		return returnValue

	@staticmethod
	def AddHostNameToSQLServerConnectionString(connectionString:str, hostName:str = None) -> str:
		if (hostName is None):
			hostName = Utilities.GetHostName()
		returnValue:str = connectionString
		returnValue += f"Workstation ID={hostName};"
		return returnValue

	@staticmethod
	def AddAppToSQLServerConnectionString(connectionString:str, appName:str= None) -> str:
		if (appName is None):
			appName = Utilities.GetPathPartsFromRight(sys.argv[0], 3, "/")
		returnValue:str = connectionString
		returnValue += f"Application Name={appName};"
		return returnValue

	@staticmethod
	def AddAppToPostgreSQLConnectionString(connectionString:str, appName:str= None) -> str:
		if (appName is None):
			appName = Utilities.GetPathPartsFromRight(sys.argv[0], 3, "/")
		returnValue:str = connectionString
		returnValue += f" application_name=\"{appName}\""
		returnValue += f" fallback_application_name=\"{appName}\""
		return returnValue

	@staticmethod
	def ParsePostgreSQLConnectionString(connectionString:str, removeSensitiveInfo:bool = False) -> dict:
		returnValue:dict = dict()
		for keyValue in connectionString.split(" "):
			keyValue = keyValue.strip()
			if (keyValue):
				keyValuePair:list = keyValue.split("=")
				key:str = ""
				value:str = ""
				if (len(keyValuePair) == 2):
					key = keyValuePair[0].strip()
					value = keyValuePair[1].strip()
					returnValue.update({key: value})
				elif (len(keyValuePair) == 1):
					key = keyValuePair[0].strip()
					value = ""
					returnValue.update({key: value})
		if (removeSensitiveInfo):
			returnValue.pop("password")
		return returnValue

	@staticmethod
	def ParseSQLServerConnectionString(connectionString:str, removeSensitiveInfo:bool = False) -> dict:
		returnValue:dict = dict()
		for keyValue in connectionString.split(";"):
			keyValue = keyValue.strip()
			if (keyValue):
				keyValuePair:list = keyValue.split("=")
				key:str = ""
				value:str = ""
				if (len(keyValuePair) == 2):
					key = keyValuePair[0].strip()
					value = keyValuePair[1].strip()
					returnValue.update({key: value})
				elif (len(keyValuePair) == 1):
					key = keyValuePair[0].strip()
					value = ""
					returnValue.update({key: value})
		if (removeSensitiveInfo):
			returnValue.pop("Password")
		return returnValue

	@staticmethod
	def ParseURIString(uriString:str, removeSensitiveInfo:bool = False) -> dict:
		returnValue:dict = dict()
		parsedSFTPURI = urlparse(uriString)
		returnValue.update({
			"hostname": parsedSFTPURI.hostname,
			"port": parsedSFTPURI.port,
			"username": parsedSFTPURI.username,
			"path": parsedSFTPURI.path
		})
		if (not removeSensitiveInfo):
			returnValue.update({ "password": parsedSFTPURI.password })
		return returnValue

	@staticmethod
	def GetRepoPath(filePath:Path = None) -> Path:
		"""Returns the first found repo path while traversing up the dir tree.
		The returned path is the parent of the first found ".git" directory.

		Keyword arguments:
		filePath -- the file (or directory) path to begin searching from. If None, then sys.argv[0] is used
		"""
		returnValue:Path = None
		if (not filePath):
			filePath = Path(sys.argv[0])
		try:
			gitDirPath = git.Repo(Path(filePath), search_parent_directories=True).git_dir
			if (gitDirPath and Path(gitDirPath).exists() and Path(gitDirPath).is_dir()):
				returnValue = Path(gitDirPath).parent
		except:
			returnValue = None
		return returnValue

	@staticmethod
	def ResolveSibling(relativePath:Path, absolutePath:Path = None) -> Path:
		"""Returns the absolute path of a relative path in relation to the absolute path.

		Keyword arguments:
		relativePath -- the path to resolve
		absolutePath -- the path to resolve by. If None, then sys.argv[0] is used
		"""
		if (absolutePath is None):
			absolutePath = Path(sys.argv[0]).absolute()
		if (absolutePath.is_file()):
			absolutePath = absolutePath.parent
		returnValue:Path = absolutePath.joinpath(relativePath).resolve()
		return returnValue

	@staticmethod
	def GetValidFileName(name:str):
		"""
		Normalizes string, converts to lowercase, removes non-alpha characters,
		and converts spaces to hyphens.
		"""
		return slugify(name)

	@staticmethod
	def GetHostName() -> str:
		return socket.gethostname().lower()

	@staticmethod
	def GetExecutingScript() -> str:
		return  str(Path(sys.argv[0]))

	@staticmethod
	def GetLineNumber() -> int:
		return inspect.currentframe().f_back.f_lineno

	@staticmethod
	def Debug(text:str, data:dict|None=None) -> None:
		"""
		Prints out the text provided, the current time in UTC, and a reference to the current file line.

		Parameters:
			text (str): The text to print to the console.
			data (dict) optional: Additional name/value paris to print to the console.

		Returns: None
		"""
		currentFrame = inspect.currentframe().f_back
		print("{0}\n  File: \"{1}\", line {2}\n  Time: {3}".format(
			text,
			currentFrame.f_code.co_filename,
			currentFrame.f_lineno,
			datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")))
		if (data is not None):
			if (len(data) > 0):
				print("  Data".ljust(20, "."))
				maxKeyLen:int = max([len(str(key)) for key in data.keys()])
				print("\n".join(["    {0}: {1}".format(
						"{0}".format(str(key)).ljust(maxKeyLen, " "),
						str(value)
					) for key, value in data.items()]))

__all__ = ["Helpers"]


from pathlib import Path
import sys
import json
from munch import Munch
import socket
from urllib.parse import urlparse
import git
import git.exc
from .ClearHelpers import ClearHelpers

class ClearConfig:
	@staticmethod
	def AppendScriptConfig(config:dict, filePath:Path, overwrite:bool = False) -> None:
		if (filePath.exists()):
			scriptConfig:dict  = json.loads(filePath.read_text())
			for key in scriptConfig.keys():
				if (key in config and overwrite):
					config[key] = scriptConfig[key]
				else:
					config.update({ key: scriptConfig[key] })

	@staticmethod
	def GetFilesInTree(beginPath:Path, endPath:Path) -> list:
		"""Returns all .env files found in the directories between beginPath and endPath.
		Note that begin path should be the top path and end path lowest path.
		The paths are search in reverse order starting at endPath going up to beginPath.
		However, the found .env files are returned in top to bottom order.

		Keyword arguments:
			beginPath -- the top path to search to
			endPath -- the bottom path to search from
		"""
		returnValue:list = []
		currentPath:Path = endPath
		if (currentPath.is_file()):
			currentPath = currentPath.parent
		while (currentPath != beginPath.parent):
			envFilePath:Path = currentPath.joinpath(".env")
			if (envFilePath.exists()):
				returnValue.append(envFilePath)
			currentPath = currentPath.parent
		returnValue.sort()
		return returnValue

	@staticmethod
	def Parse(contents:str) -> dict:
		"""Returns a dict containing any name=value pairs found in the string.
		Comment (#) lines are ignored.
		Comments should not be included inline with name=value pairs.

		Keyword arguments:
			contents -- the string to parse
		"""
		if (contents is not None):
			lines:list = contents.splitlines()
			if (len(lines) > 0):
				returnValue = {}
				for line in lines:
					line = line.strip()
					if (not line.startswith("#") and len(line) > 0):
						#Find first eqaul sign
						firstEqual:int = line.find("=")

						#Split into name/value
						name = line[:firstEqual].strip()
						value = line[firstEqual+1:].strip()

						#Strip leading and trailing double quotes from name
						if name.startswith('"') and name.endswith('"'):
							name = name[1:-1]

						#Strip leading and trailing single quotes from name
						if name.startswith("'") and name.endswith("'"):
							name = name[1:-1]

						#Strip leading and trailing double quotes from value
						if value.startswith('"') and value.endswith('"'):
							value = value[1:-1]

						#Strip leading and trailing single quotes from value
						if value.startswith("'") and value.endswith("'"):
							value = value[1:-1]

						#Attempt to resolve boolean values
						if (value in ["True", "TRUE", "true"]):
							returnValue.update({name: True})
						elif (value in ["False", "FALSE", "false"]):
							returnValue.update({name: False})
						#If not boolean, then
						else:
							returnValue.update({name: value})
		return returnValue

	@staticmethod
	def OpenFilePath(filePath:Path, resolveSmartSettings:bool = False) -> dict:
		"""Returns a dict containing any name=value pairs found in the specified file.
		Comment (#) lines are ignored.
		Comments should not be included inline with name=value pairs.

		Keyword arguments:
			filePath -- the path to the file to parse
			resolveSmartSettings -- Alters settings as follows and returns appropriate type
				If key begins with PATH_, converts relative paths to absolute paths
				If key begins with SQLCONSTRING_, appends application name and workstation name to connection strings
				If key begins with POSTGRECONSTRING_, appends application name to connection strings
				If key begins with LIST_, converts value to a list, splitting on the pipe (|) character
		"""
		returnValue:dict = None
		fileContents:str = None
		with open(filePath, "r") as file:
			fileContents = file.read()
		if (fileContents is not None):
			returnValue = ClearConfig.Parse(fileContents)
			for key in returnValue:
				if (resolveSmartSettings):
					if (key.startswith("PATH_")):
						returnValue.update({key: ClearHelpers.ResolveSibling(Path(returnValue[key]), filePath)})
					if (key.startswith("SQLCONSTRING_")):
						returnValue.update({key: 
							ClearHelpers.AddAppToSQLServerConnectionString(
								ClearHelpers.AddHostNameToSQLServerConnectionString(
									str(returnValue[key])
						))})
					if (key.startswith("POSTGRECONSTRING_")):
						returnValue.update({key: ClearHelpers.AddAppToPostgreSQLConnectionString(str(returnValue[key]))})
					if (key.startswith("LIST_")):
						returnValue.update({key: str(returnValue[key]).split("|")})
		return returnValue

	@staticmethod
	def OpenFilesInTree(beginPath:Path, endPath:Path, resolveSmartSettings:bool = False) -> dict:
		"""Returns a dict containing any name=value pairs found in any .env files located in the directories between beginPath and endPath.
		Note that begin path should be the top path and end path lowest path.
		The paths are search in reverse order starting at endPath going up to beginPath.
		However, the found .env files are processed in top to bottom order.
		Duplicate key names found are overriden by the value in the lower level .env.

		Keyword arguments:
			beginPath -- the top path to search to
			endPath -- the bottom path to search from
			resolveSmartSettings -- Alters settings as follows and returns appropriate type
				If key begins with PATH_, converts relative paths to absolute paths
				If key begins with SQLCONSTRING_, appends application name and workstation name to connection strings
				If key begins with POSTGRECONSTRING_, appends application name to connection strings
				If key begins with LIST_, converts value to a list, splitting on the pipe (|) character
		"""
		returnValue:dict = {}
		if (beginPath is None or endPath is None):
			raise ValueError("Paths cannot be None")
		envFilePathList:list = ClearConfig.GetFilesInTree(beginPath, endPath)
		for envFilePath in envFilePathList:
			returnValue.update(ClearConfig.OpenFilePath(envFilePath, resolveSmartSettings))
		return dict(sorted(returnValue.items()))

	@staticmethod
	def OpenRepoEnv(resolveSmartSettings:bool = False, jsonScriptConfigPath:Path|str|None=None) -> dict:
		"""Returns a dict containing any name=value pairs found in any .env files located in the directories between the first found git repo directory and sys.argv[0] file.
		The paths are search in reverse order starting at sys.argv[0] path going up to repo path.
		However, the found .env files are processed in top to bottom order.
		Duplicate key names found are overriden by the value in the lower level .env.

		Keyword arguments:
			resolveSmartSettings -- Alters settings as follows and returns appropriate type
				If key begins with PATH_, converts relative paths to absolute paths
				If key begins with SQLCONSTRING_, appends application name and workstation name to connection strings
				If key begins with POSTGRECONSTRING_, appends application name to connection strings
				If key begins with LIST_, converts value to a list, splitting on the pipe (|) character
			jsonScriptConfigPath -- If specified, adds data from JSON file to config.
		"""
		returnValue:dict = None
		returnValue = ClearConfig.OpenFilesInTree(ClearHelpers.GetRepoPath(), Path(sys.argv[0]), resolveSmartSettings)
		if (jsonScriptConfigPath is not None):
			if (isinstance(jsonScriptConfigPath, str)):
				scriptConfigPath = Path(jsonScriptConfigPath)
				if (jsonScriptConfigPath.exists()):
					if (returnValue is None):
						returnValue = dict()
					ClearConfig.AppendScriptConfig(returnValue, jsonScriptConfigPath, True)
		return returnValue

	@staticmethod
	def OpenFile(filePath:str = None) -> dict:
		returnValue:dict = None
		fileContents:str = None
		if (not filePath):
			filePath = str(Path(sys.argv[0]).parent.joinpath(".env"))
		if (not Path(filePath).is_file()):
			raise FileNotFoundError(filePath)
		else:
			with open(filePath, "r") as file:
				fileContents = file.read()
		if (fileContents is not None):
			returnValue = ClearConfig.Parse(fileContents)
		return returnValue

__all__ = ["ClearConfig"]
